from onegov.user.model import User
from onegov.user.errors import UnknownUserError
from sqlalchemy import sql


class UserCollection(object):
    """ Manages a list of users.

    Use it like this::

        from onegov.user import UserCollection
        users = UserCollection(session)

    """

    def __init__(self, session):
        self.session = session

    def query(self):
        """ Returns a query using :class:`onegov.user.model.User`. """
        return self.session.query(User)

    def add(self, username, password, role, data=None):
        """ Add a user to the collection.

            :username:
                Must be a unique string and may not be empty.

            :password:
                Must be a non-empty string.

            :role:
                Must be a non-empty string. The meaning of roles is not defined
                in `onegov.user` instead the role is usually defined by
                :mod:`onegov.core.security.roles`.
        """
        assert username and password and role

        user = User(username=username, password=password, role=role, data=data)

        self.session.add(user)
        self.session.flush()

        return user

    def exists(self, username):
        """ Returns True if the given username exists.

        This function does not actually load a user, so it is the quickest
        way to find out if a user exists. It should be used if you don't
        care about finding out anything about the user.

        """
        query = self.session.query(sql.exists().where(
            User.username == username))

        return query.scalar()

    def by_username(self, username):
        """ Returns the user by username. """
        return self.query().filter(User.username == username).first()

    def by_username_and_password(self, username, password):
        """ Returns the user by username and password.

        Note that although the password can be empty on the user, this function
        will not query for empty passwords as an added security measure.

        Apart from that everything is fair game though, as it is not the job
        of onegov.user to enforce a password policy.

        """
        user = self.by_username(username)

        if user and password and user.is_matching_password(password):
            return user
        else:
            return None

    def delete(self, username):
        """ Deletes the user if it exists.

        If the user does not exist, an
        :class:`onegov.user.errors.UnknownUserError` is raised.

        """
        user = self.by_username(username)

        if not user:
            raise UnknownUserError("user {} does not exist".format(username))

        self.session.delete(user)
        self.session.flush()



import morepath
import transaction

from more.itsdangerous import IdentityPolicy
from onegov.user import Auth, UserCollection
from webtest import TestApp as Client


def test_auth_login(session):
    UserCollection(session).add('AzureDiamond', 'hunter2', 'irc-user')
    auth = Auth(session=session, application_id='my-app')

    assert not auth.login(username='AzureDiamond', password='hunter1')
    assert not auth.login(username='AzureDiamonb', password='hunter2')

    identity = auth.login(username='AzureDiamond', password='hunter2')
    assert identity.userid == 'AzureDiamond'
    assert identity.role == 'irc-user'
    assert identity.application_id == 'my-app'


def test_auth_logging(session, capturelog):
    UserCollection(session).add('AzureDiamond', 'hunter2', 'irc-user')
    auth = Auth(session=session, application_id='my-app')

    auth.login(username='AzureDiamond', password='hunter1')
    assert capturelog.records()[0].message \
        == "Failed login by unknown (AzureDiamond)"

    auth.login(username='AzureDiamond', password='hunter1', client='127.0.0.1')
    assert capturelog.records()[1].message \
        == "Failed login by 127.0.0.1 (AzureDiamond)"

    auth.login(username='AzureDiamond', password='hunter2', client='127.0.0.1')
    assert capturelog.records()[2].message \
        == "Successful login by 127.0.0.1 (AzureDiamond)"


def test_auth_integration(session):

    config = morepath.setup()

    class App(morepath.App):
        testing_config = config

    @App.identity_policy()
    def get_identity_policy():
        return IdentityPolicy()

    @App.path(path='/auth', model=Auth)
    def get_auth():
        return Auth(session, application_id='my_app', to='https://abc.xyz/go')

    @App.view(model=Auth)
    def view_auth(self, request):
        return self.login_to(
            request.params.get('username'),
            request.params.get('password'),
            request
        ) or 'Error'

    @App.view(model=Auth, name='logout')
    def view_logout(self, request):
        return self.logout_to(request)

    config.commit()

    UserCollection(session).add('AzureDiamond', 'hunter2', 'irc-user')
    transaction.commit()

    client = Client(App())

    response = client.get('/auth?username=AzureDiamond&password=hunter1')
    assert response.text == 'Error'

    response = client.get('/auth?username=AzureDiamond&password=hunter2')
    assert response.status_code == 302
    assert response.location == 'http://localhost/go'
    assert response.headers['Set-Cookie'].startswith('userid=AzureDiamond')

    response = client.get('/auth/logout')
    assert response.status_code == 302
    assert response.location == 'http://localhost/go'
    assert response.headers['Set-Cookie'].startswith('userid=;')



import morepath
from onegov.core.utils import relative_url
from onegov.user import log
from onegov.user.collection import UserCollection


class Auth(object):
    """ Defines a model for authentication methods like login/logout.
    Applications should use this model to implement authentication views.

    """

    identity_class = morepath.Identity

    def __init__(self, session, application_id, to='/'):
        assert application_id  # may not be empty!

        self.session = session
        self.application_id = application_id

        # never redirect to an external page, this might potentially be used
        # to trick the user into thinking he's on our page after entering his
        # password and being redirected to a phising site.
        self.to = relative_url(to)

    @classmethod
    def from_app(cls, app, to='/'):
        return cls(app.session(), app.application_id, to)

    @classmethod
    def from_request(cls, request, to='/'):
        return cls.from_app(request.app)

    @property
    def users(self):
        return UserCollection(self.session)

    def login(self, username, password, client='unknown'):
        """ Takes the given username and password and matches them against
        the users collection.

        :meth:`login_to` should be the preferred method in most cases, as it
        sets up automatic logging of login attempts and provides optional
        login/logout messages.

        :return: An identity bound to the ``application_id`` if the username
        and password match.

        """

        user = self.users.by_username_and_password(username, password)

        if user is None:
            log.info("Failed login by {} ({})".format(client, username))
            return None

        log.info("Successful login by {} ({})".format(client, username))

        return self.identity_class(
            userid=user.username,
            role=user.role,
            application_id=self.application_id
        )

    def login_to(self, username, password, request):
        """ Matches the username and password against the users collection,
        just like :meth:`login`. Unlike said method it returns no identity
        however, instead a redirect response is returned which will remember
        the identity.

        :return: A redirect response to ``self.to`` with the identity
        remembered as a cookie. If not successful, None is returned.

        """

        identity = self.login(username, password, request.client_addr)

        if identity is None:
            return None

        response = morepath.redirect(self.to)
        morepath.remember_identity(response, request, identity)

        return response

    def logout_to(self, request):
        """ Logs the current user out and redirects to ``self.to``.

        :return: A response redirecting to ``self.to`` with the identity
        forgotten.

        """

        response = morepath.redirect(self.to)
        morepath.forget_identity(response, request)

        return response



""" Provides commands used to manage users.

Examples:

To add a user called ``admin@example.org`` to the ``towns-govikon`` schema,
with the role 'admin'::

    dsn=postgres://user:password@localhost:5432/database
    schema=towns-govikon
    onegov-user --dsn $dsn --schema $schema add admin admin@example.org

This command will ask for a password if none was provided with ``--password``.

To delete a user::

    dsn=postgres://user:password@localhost:5432/database
    schema=towns-govikon
    onegov-user --dsn $dsn --schema $schema delete admin@example.org

To check if a user exists::

    dsn=postgres://user:password@localhost:5432/database
    schema=towns-govikon
    onegov-user --dsn $dsn --schema $schema exists admin@example.org

To change a users password::

    dsn=postgres://user:password@localhost:5432/database
    schema=towns-govikon
    onegov-user --dsn $dsn --schema $schema change-password admin@example.org

This command will also ask for a password if none was provided with
``--password``.

"""

import click
import sys
import transaction

from getpass import getpass
from onegov.user import UserCollection
from onegov.core.orm import Base, SessionManager


@click.group()
@click.option('--dsn', help="Postgresql connection string")
@click.option('--schema', help="Schema to use")
@click.pass_context
def cli(ctx, dsn, schema):
    ctx.obj = {}

    mgr = SessionManager(dsn, base=Base)
    mgr.set_current_schema(schema)

    ctx.obj['schema'] = schema
    ctx.obj['session'] = mgr.session()


@cli.command()
@click.argument('role')
@click.argument('username')
@click.option('--password', help="Password to give the user", default=None)
@click.pass_context
def add(ctx, role, username, password):
    """ Adds a user with the given name to the database. """

    session = ctx.obj['session']
    users = UserCollection(session)

    if users.exists(username):
        click.secho("The user {} already exists".format(username), fg='red')
        sys.exit(1)

    password = password or getpass("Enter password: ")

    users.add(username, password, role)
    transaction.commit()

    click.secho("The user {} was added".format(username), fg='green')


@cli.command()
@click.argument('username')
@click.pass_context
def delete(ctx, username):
    """ Removes the given user from the database. """

    session = ctx.obj['session']
    users = UserCollection(session)

    if not users.exists(username):
        click.secho("The user {} does not exist".format(username), fg='red')
        sys.exit(1)

    users.delete(username)
    transaction.commit()

    click.secho("The user {} was deleted".format(username), fg='green')


@cli.command()
@click.argument('username')
@click.pass_context
def exists(ctx, username):
    """ Returns 0 if the user exists, 1 if it doesn't. """

    session = ctx.obj['session']
    users = UserCollection(session)

    if not users.exists(username):
        click.secho("The user {} does not exist".format(username), fg='red')
        sys.exit(1)
    else:
        click.secho("The user {} exists".format(username), fg='green')
        sys.exit(0)


@cli.command(name='change-password')
@click.argument('username')
@click.option('--password', help="Password to use", default=None)
@click.pass_context
def change_password(ctx, username, password):
    """ Changes the password of the given username. """

    session = ctx.obj['session']
    users = UserCollection(session)

    if not users.exists(username):
        click.secho("The user {} does not exist".format(username), fg='red')
        sys.exit(1)

    password = password or getpass("Enter password: ")

    user = users.by_username(username)
    user.password = password

    transaction.commit()

    click.secho("The password for {} was changed".format(username), fg='green')



from onegov.form import Form
from onegov.user import _
from wtforms import StringField, PasswordField, validators


class LoginForm(Form):
    """ A generic login form for onegov.user """

    username = StringField(
        label=_("Username/E-Mail Address"),
        validators=[validators.InputRequired(), validators.Email()]
    )
    password = PasswordField(
        label=_("Password"),
        validators=[validators.InputRequired()]
    )

    @property
    def login_data(self):
        """ Returns the data required to be passed to the
        :class:`onegov.user.auth.Auth` methods.

        """

        return {
            'username': self.username.data,
            'password': self.password.data,
        }



from click.testing import CliRunner
from onegov.user.cli import cli


def test_cli(postgres_dsn):

    runner = CliRunner()
    result = runner.invoke(cli, [
        '--dsn', postgres_dsn,
        '--schema', 'test-add-user',
        'add', 'admin', 'admin@example.org',
        '--password', 'hunter2'
    ])

    assert result.exit_code == 0
    assert 'The user admin@example.org was added' in result.output

    result = runner.invoke(cli, [
        '--dsn', postgres_dsn,
        '--schema', 'test-add-user',
        'add', 'admin', 'admin@example.org',
        '--password', 'hunter2'
    ])

    assert result.exit_code == 1
    assert 'The user admin@example.org already exists' in result.output

    result = runner.invoke(cli, [
        '--dsn', postgres_dsn,
        '--schema', 'test-add-user',
        'exists', 'admin@example.org'
    ])

    assert result.exit_code == 0
    assert 'The user admin@example.org exists' in result.output

    result = runner.invoke(cli, [
        '--dsn', postgres_dsn,
        '--schema', 'test-add-user',
        'change-password', 'admin@example.org',
        '--password', 'hunter2'
    ])

    assert result.exit_code == 0
    assert 'The password for admin@example.org was changed' in result.output

    result = runner.invoke(cli, [
        '--dsn', postgres_dsn,
        '--schema', 'test-add-user',
        'delete', 'admin@example.org'
    ])

    assert result.exit_code == 0
    assert 'The user admin@example.org was deleted' in result.output

    result = runner.invoke(cli, [
        '--dsn', postgres_dsn,
        '--schema', 'test-add-user',
        'delete', 'admin@example.org'
    ])

    assert result.exit_code == 1
    assert 'The user admin@example.org does not exist' in result.output

    result = runner.invoke(cli, [
        '--dsn', postgres_dsn,
        '--schema', 'test-add-user',
        'exists', 'admin@example.org'
    ])

    assert result.exit_code == 1
    assert 'The user admin@example.org does not exist' in result.output

    result = runner.invoke(cli, [
        '--dsn', postgres_dsn,
        '--schema', 'test-add-user',
        'change-password', 'admin@example.org',
        '--password', 'hunter2'
    ])

    assert result.exit_code == 1
    assert 'The user admin@example.org does not exist' in result.output
